# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from ..collector import DueCreditCollector, Citation
from .test_collector import _sample_bibtex, _sample_doi
from ..entries import BibTeX, DueCreditEntry, Doi
from ..io import TextOutput, PickleOutput, import_doi, EnumeratedEntries, get_text_rendering
from nose.tools import assert_equal, assert_is_instance, assert_raises, \
    assert_true, assert_false
from six.moves import StringIO
from six import text_type

from mock import patch

import random
import re
import sys
import pickle
import tempfile
from .test_collector import _sample_bibtex, _sample_bibtex2

try:
    import vcr

    @vcr.use_cassette()
    def test_import_doi():
        doi_good = '10.1038/nrd842'
        assert_is_instance(import_doi(doi_good), text_type)

        doi_bad = 'fasljfdldaksj'
        assert_raises(ValueError, import_doi, doi_bad)

except ImportError:
    # no vcr, and that is in 2015!
    pass


def test_pickleoutput():
    #entry = BibTeX('@article{XXX0, ...}')
    entry = BibTeX("@article{Atkins_2002,\n"
                   "title=title,\n"
                   "volume=1, \n"
                   "url=http://dx.doi.org/10.1038/nrd842, \n"
                   "DOI=10.1038/nrd842, \n"
                   "number=7, \n"
                   "journal={Nat. Rev. Drug Disc.}, \n"
                   "publisher={Nature Publishing Group}, \n"
                   "author={Atkins, Joshua H. and Gershell, Leland J.}, \n"
                   "year={2002}, \n"
                   "month={Jul}, \n"
                   "pages={491--492}\n}")
    collector_ = DueCreditCollector()
    collector_.add(entry)
    collector_.cite(entry, path='module')

    # test it doesn't puke with an empty collector
    collectors = [collector_, DueCreditCollector()]

    for collector in collectors:
        with tempfile.NamedTemporaryFile() as fn:
            pickler = PickleOutput(collector, fn=fn.name)
            assert_equal(pickler.fn, fn.name)
            assert_equal(pickler.dump(), None)
            collector_loaded = pickle.load(fn)

            assert_equal(collector.citations.keys(),
                         collector_loaded.citations.keys())
            # TODO: implement comparison of citations
            assert_equal(collector._entries.keys(),
                         collector_loaded._entries.keys())

def test_text_output():
    entry = BibTeX(_sample_bibtex)
    collector = DueCreditCollector()
    collector.cite(entry, path='module')

    strio = StringIO()
    TextOutput(strio, collector).dump(tags=['*'])
    value = strio.getvalue()
    assert_true("Halchenko, Y.O." in value, msg="value was %s" % value)
    assert_true(value.strip().endswith("Frontiers in Neuroinformatics, 6(22)."))


def test_text_output_dump_formatting():
    due = DueCreditCollector()

    # XXX: atm just to see if it spits out stuff
    @due.dcite(BibTeX(_sample_bibtex), description='solution to life',
               path='mymodule', version='0.0.16')
    def mymodule(arg1, kwarg2="blah"):
        """docstring"""
        assert_equal(arg1, "magical")
        assert_equal(kwarg2, 1)

        @due.dcite(BibTeX(_sample_bibtex2), description='solution to life',
                   path='mymodule:myfunction')
        def myfunction(arg42):
            pass

        myfunction('argh')
        return "load"

    # check we don't have anything output
    strio = StringIO()
    TextOutput(strio, due).dump(tags=['*'])
    value = strio.getvalue()
    assert_true('0 modules cited' in value, msg='value was {0}'.format(value))
    assert_true('0 functions cited' in value,
                msg='value was {0}'.format(value))

    # now we call it -- check it prints stuff
    mymodule('magical', kwarg2=1)
    TextOutput(strio, due).dump(tags=['*'])
    value = strio.getvalue()
    assert_true('1 packages cited' in value, msg='value was {0}'.format(value))
    assert_true('1 functions cited' in value, msg='value was {0}'.format(value))
    assert_true('(v 0.0.16)' in value,
                msg='value was {0}'.format(value))
    assert_equal(len(value.split('\n')), 21, msg='value was {0}'.format(value))

    # test we get the reference numbering right
    samples_bibtex = [_generate_sample_bibtex() for x in range(5)]
    # this sucks but at the moment it's the only way to have multiple
    # references for a function

    @due.dcite(BibTeX(samples_bibtex[0]), description='another solution',
               path='myothermodule', version='0.0.666')
    def myothermodule(arg1, kwarg2="blah"):
        """docstring"""
        assert_equal(arg1, "magical")
        assert_equal(kwarg2, 1)

        @due.dcite(BibTeX(samples_bibtex[1]), description='solution to life',
                   path='myothermodule:myotherfunction')
        @due.dcite(BibTeX(samples_bibtex[2]), description='solution to life',
                   path='myothermodule:myotherfunction')
        @due.dcite(BibTeX(samples_bibtex[3]), description='solution to life',
                   path='myothermodule:myotherfunction')
        @due.dcite(BibTeX(samples_bibtex[4]), description='solution to life',
                   path='myothermodule:myotherfunction')
        @due.dcite(BibTeX(_sample_bibtex2), description='solution to life',
                   path='myothermodule:myotherfunction')
        def myotherfunction(arg42):
            pass

        myotherfunction('argh')
        return "load"

    myothermodule('magical', kwarg2=1)
    strio = StringIO()
    TextOutput(strio, due).dump(tags=['*'])
    value = strio.getvalue()
    lines = value.split('\n')

    citation_numbers = []
    reference_numbers = []
    references = []
    for line in lines:
        match_citation = re.search('\[([0-9, ]+)\]$', line)
        match_reference = re.search('^\[([0-9])\]', line)
        if match_citation:
            citation_numbers.extend(match_citation.group(1).split(', '))
        elif match_reference:
            reference_numbers.append(match_reference.group(1))
            references.append(line.replace(match_reference.group(), ""))

    assert_equal(set(citation_numbers), set(reference_numbers))
    assert_equal(len(set(references)), len(set(citation_numbers)))
    assert_equal(len(citation_numbers), 8)
    # verify that we have returned to previous state of filters
    import warnings
    assert_true(('ignore', None, UserWarning, None, 0) not in warnings.filters)

def _generate_sample_bibtex():
    """
    Generate a random sample bibtex to test multiple references
    """
    letters = 'abcdefghilmnopqrstuvxz'
    numbers = '0123456789'
    letters_numbers = letters + letters.upper() + numbers
    letters_numbers_spaces = letters_numbers + ' '

    key = "".join(random.sample(letters_numbers, 7))
    title = "".join(random.sample(letters_numbers_spaces, 20))
    journal = "".join(random.sample(letters_numbers_spaces, 20))
    publisher = "".join(random.sample(letters_numbers_spaces, 10))
    author = "".join(random.sample(letters, 6)) + ', ' + \
             "".join(random.sample(letters, 4))
    year = "".join(random.sample(numbers, 4))

    elements = [('title', title), ('journal', journal),
                ('publisher', publisher), ('author', author),
                ('year', year)]

    sample_bibtex = "@ARTICLE{%s,\n" % key
    for string, value in elements:
        sample_bibtex += "%s={%s},\n" % (string, value)
    sample_bibtex += "}"
    return sample_bibtex

def test_enumeratedentries():
    enumentries = EnumeratedEntries()
    assert_false(enumentries)

    # add some entries
    entries = [('ciao', 1), ('miao', 2), ('bau', 3)]
    for entry, _ in entries:
        enumentries.add(entry)

    assert_equal(len(enumentries), 3)

    for entry, nr in entries:
        assert_equal(nr, enumentries[entry])
        assert_equal(entry, enumentries.fromrefnr(nr))

    assert_raises(KeyError, enumentries.__getitem__, 'boh')
    assert_raises(KeyError, enumentries.fromrefnr, 666)

    assert_equal(entries, sorted(enumentries, key=lambda x: x[1]))

@patch('duecredit.io.get_bibtex_rendering')
@patch('duecredit.io.format_bibtex')
def test_get_text_rendering(mock_format_bibtex, mock_get_bibtex_rendering):
    # mock get_bibtex_rendering to return the same bibtex entry
    sample_bibtex = BibTeX(_sample_bibtex)
    mock_get_bibtex_rendering.return_value = sample_bibtex

    # test if bibtex type is passed
    citation_bibtex = Citation(sample_bibtex, path='mypath')
    bibtex_output = get_text_rendering(citation_bibtex)
    mock_format_bibtex.assert_called_with(citation_bibtex.entry, style='harvard1')
    mock_format_bibtex.reset_mock()

    # test if doi type is passed
    citation_doi = Citation(Doi(_sample_doi), path='mypath')
    doi_output = get_text_rendering(citation_doi)
    mock_format_bibtex.assert_called_with(citation_bibtex.entry, style='harvard1')

    assert_equal(bibtex_output, doi_output)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""
Automatic injection of bibliography entries for psychopy module
"""


from ..entries import Doi, BibTeX, Url

# If defined, would determine from which to which version of the corresponding
# module to care about
min_version = None
max_version = None


def inject(injector):
    injector.add('psychopy', None, Doi('doi:10.1016/j.jneumeth.2006.11.017'),
                 description="PsychoPy -- Psychophysics software in Python.",
                 tags=['implementation'])
    
    injector.add('psychopy', None, Doi('10.3389/neuro.11.010.2008'),
                 description="Generating stimuli for neuroscience using PsychoPy.",
                 tags=['implementation'])



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from duecredit.collector import DueCreditCollector
from duecredit.stub import InactiveDueCreditCollector
from duecredit.entries import BibTeX, Doi

from nose.tools import assert_equal
from nose.tools import assert_in
from nose import SkipTest

def _test_api(due):
    # add references
    due.add(BibTeX('@article{XXX00, ...}'))
    # could even be by DOI -- we need to fetch and cache those
    due.add(Doi("xxx.yyy/zzz.1", key="XXX01"))

    # and/or load multiple from a file
    due.load('/home/siiioul/deep/good_intentions.bib')

    # Cite entire module
    due.cite('XXX00', description="Answers to existential questions", path="module")
    # Cita some method within some submodule
    due.cite('XXX01', description="More answers to existential questions",
             path="module.submodule:class1.whoknowswhat2.func1")

    # dcite  for decorator cite
    # cite specific functionality if/when it gets called up
    @due.dcite('XXX00', description="Provides an answer for meaningless existence")
    def purpose_of_life():
        return None

    class Child(object):
         # Conception process is usually way too easy to be referenced
         def __init__(self):
             pass

         # including functionality within/by the methods
         @due.dcite('XXX00')
         def birth(self, gender):
             return "Rachel was born"

    kid = Child()
    kid.birth("female")


def test_api():
    yield _test_api, DueCreditCollector()
    yield _test_api, InactiveDueCreditCollector()

import os
import sys
from os.path import dirname, join as pathjoin, pardir, normpath
from mock import patch
from subprocess import Popen, PIPE

badlxml_path = pathjoin(dirname(__file__), 'envs', 'nolxml')
stubbed_script = pathjoin(dirname(__file__), 'envs', 'stubbed', 'script.py')

def run_python_command(cmd=None, script=None):
    """Just a tiny helper which runs command and returns exit code, stdout, stderr"""
    assert(bool(cmd) != bool(script))  # one or another, not both
    args = ['-c', cmd] if cmd else [script]
    python = Popen([sys.executable] + args, stdout=PIPE, stderr=PIPE)
    ret = python.wait()
    return ret, python.stdout.read().decode(), python.stderr.read().decode()

mock_env_nolxml = {'PYTHONPATH': "%s:%s" % (badlxml_path, os.environ.get('PYTHONPATH', ''))}

# Since duecredit and possibly lxml already loaded, let's just test
# ability to import in absence of lxml via external call to python
def test_noincorrect_import_if_no_lxml():
    with patch.dict(os.environ, mock_env_nolxml):
        # make sure out mocking works here
        ret, out, err = run_python_command('import lxml')
        assert_equal(ret, 1)
        assert_in('ImportError', err)
        #
        # make sure out mocking works here
        ret, out, err = run_python_command('import duecredit')
        assert_equal(err, '')
        assert_equal(out, '')
        assert_equal(ret, 0)

def check_noincorrect_import_if_no_lxml_numpy(kwargs, env):
    # Now make sure that we would not crash entire process at the end when unable to
    # produce sensible output when we have something to cite
    # we do inject for numpy
    try:
        import numpy
    except ImportError:
        raise SkipTest("We need to have numpy to test correct operation")

    mock_env_nolxml_ = mock_env_nolxml.copy()
    mock_env_nolxml_.update(env)

    with patch.dict(os.environ, mock_env_nolxml_):
        ret, out, err = run_python_command(**kwargs)
        assert_equal(err, '')
        if os.environ.get('DUECREDIT_ENABLE', False):  # we enabled duecredit
            assert_in('For formatted output we need citeproc', out)
            assert_in('done123', out)
        elif os.environ.get('DUECREDIT_TEST_EARLY_IMPORT_ERROR'):
            assert_in('ImportError', out)
            assert_in('DUECREDIT_TEST_EARLY_IMPORT_ERROR', out)
            assert_in('done123', out)
        else:
            assert_equal('done123\n', out)
        assert_equal(ret, 0)  # but we must not fail overall regardless


def test_noincorrect_import_if_no_lxml_numpy():
    for kwargs in (
        # direct command to evaluate
        {'cmd': 'import duecredit; import numpy as np; print("done123")'},
        # script with decorated funcs etc -- should be importable
        {'script': stubbed_script}
    ):
        yield check_noincorrect_import_if_no_lxml_numpy, kwargs, {}
        yield check_noincorrect_import_if_no_lxml_numpy, kwargs, {'DUECREDIT_ENABLE': 'yes'}
        yield check_noincorrect_import_if_no_lxml_numpy, kwargs, {'DUECREDIT_TEST_EARLY_IMPORT_ERROR': 'yes'}



if __name__ == '__main__':
    from duecredit import due
    _test_api(due)


# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.   Originates from datalad package distributed
#   under MIT license
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

import sys
from mock import patch

from ..utils import is_interactive
from nose.tools import assert_false, assert_true

def test_is_interactive_crippled_stdout():
    class mocked_out(object):
        """the one which has no isatty
        """
        def write(self, *args, **kwargs):
            pass

    class mocked_isatty(mocked_out):
        def isatty(self):
            return True

    for inout in ('in', 'out', 'err'):
        with patch('sys.std%s' % inout, mocked_out()):
            assert_false(is_interactive())

    # just for paranoids
    with patch('sys.stdin', mocked_isatty()), \
            patch('sys.stdout', mocked_isatty()), \
            patch('sys.stderr', mocked_isatty()):
        assert_true(is_interactive())


# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

from mock import patch
import atexit

from nose import SkipTest

from ..injections.injector import DueCreditInjector
from ..dueswitch import due

@patch.object(DueCreditInjector, 'activate')
@patch.object(atexit, 'register')
def test_dueswitch_activate(mock_register, mock_activate):
    was_active = due.active
    # atexit.register(crap)
    # injector.activate()
    due.activate()
    if was_active:
        # we can only test that mocked methods do not invoked second time
        mock_activate.assert_not_called()
        mock_register.assert_not_called()
        raise SkipTest("due is already active, can't test more at this point")
    # was not active, so should have called activate of the injector class
    mock_activate.assert_called_once_with()
    mock_register.assert_called_once_with(due._dump_collector_summary)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Provides an adapter to switch between two (active, inactive) collectors
"""

import os
import atexit

from .log import lgr
from .utils import never_fail

def _get_duecredit_enable():
    env_enable = os.environ.get('DUECREDIT_ENABLE', 'no')
    if not env_enable.lower() in ('0', '1', 'yes', 'no', 'true', 'false'):
        lgr.warning("Misunderstood value %s for DUECREDIT_ENABLE. "
                    "Use 'yes' or 'no', or '0' or '1'")
    return env_enable.lower() in ('1', 'yes', 'true')

@never_fail
def _get_inactive_due():
    # keeping duplicate but separate so later we could even place it into a separate
    # submodule to possibly minimize startup time impact even more
    from .collector import InactiveDueCreditCollector
    return InactiveDueCreditCollector()

@never_fail
def _get_active_due():
    from .config import CACHE_DIR, DUECREDIT_FILE
    from duecredit.collector import CollectorSummary, DueCreditCollector
    from .io import load_due
    import atexit
    # where to cache bibtex entries
    if not os.path.exists(CACHE_DIR):
        os.makedirs(CACHE_DIR)

    # TODO:  this needs to move to atexit handling, that we load previous
    # one and them merge with new ones.  Informative bits could be -- how
    # many new citations we got
    if os.path.exists(DUECREDIT_FILE):
        try:
            due_ = load_due(DUECREDIT_FILE)
        except Exception as e:
            lgr.warning("Failed to load previously collected %s. "
                        "DueCredit will not be active for this session."
                        % DUECREDIT_FILE)
            return _get_inactive_due()
    else:
        due_ = DueCreditCollector()

    return due_


class DueSwitch(object):
    """Adapter between two types of collectors -- Inactive and Active

    Once activated though, cannot be fully deactivated since it would inject
    duecredit decorators and register an event atexit.
    """
    def __init__(self, inactive, active, activate=False):
        self.__active = None
        self.__collectors = {False: inactive, True: active}
        self.__activations_done = False
        self.activate(activate)

    @property
    def active(self):
        return self.__active

    @never_fail
    def _dump_collector_summary(self):
        from duecredit.collector import CollectorSummary
        due_summary = CollectorSummary(self.__collectors[True])
        due_summary.dump()

    def __prepare_exit_and_injections(self):
        # Wrapper to create and dump summary... passing method doesn't work:
        #  probably removes instance too early

        atexit.register(self._dump_collector_summary)

        # Deal with injector
        from .injections import DueCreditInjector
        injector = DueCreditInjector(collector=self.__collectors[True])
        injector.activate()

    @never_fail
    def activate(self, activate=True):
        # 1st step -- if activating/deactivating switch between the two collectors
        if self.__active is not activate:
            # we need to switch the state
            is_public = lambda x: not x.startswith('_')
            # Clean up current bindings first
            for k in filter(is_public, dir(self)):
                if k not in ('activate', 'active'):
                    delattr(self, k)

            new_due = self.__collectors[activate]
            for k in filter(is_public, dir(new_due)):
                setattr(self, k, getattr(new_due, k))

            self.__active = activate

        # 2nd -- if activating, we might still need to have activations done
        if activate and not self.__activations_done:
            try:
                self.__prepare_exit_and_injections()
            except Exception as e:
                lgr.error("Failed to prepare injections etc: %s" % str(e))
            finally:
                self.__activations_done = True


due = DueSwitch(_get_inactive_due(), _get_active_due(), _get_duecredit_enable())


# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

# Just for testing of robust operation
import os
if 'DUECREDIT_TEST_EARLY_IMPORT_ERROR' in os.environ.keys():
    raise ImportError("DUECREDIT_TEST_EARLY_IMPORT_ERROR")

import time
from collections import defaultdict, Iterator
import copy
from os.path import dirname, exists
import pickle
import requests
import tempfile
from six import PY2, itervalues, iteritems
import warnings

from .config import CACHE_DIR, DUECREDIT_FILE
from .entries import BibTeX, Doi
from .log import lgr

def get_doi_cache_file(doi):
    return os.path.join(CACHE_DIR, doi)


def import_doi(doi):
    cached = get_doi_cache_file(doi)

    if exists(cached):
        with open(cached) as f:
            doi = f.read()
            if PY2:
                return doi.decode('utf-8')
            return doi

    # else -- fetch it
    headers = {'Accept': 'text/bibliography; style=bibtex'}
    url = 'http://dx.doi.org/' + doi
    retries = 10
    while retries > 0:
        r = requests.get(url, headers=headers)
        r.encoding = 'UTF-8'
        bibtex = r.text.strip()
        if bibtex.startswith('@'):
            # no more retries necessary
            break
        lgr.warning("Failed to obtain bibtex from doi.org, retrying...")
        time.sleep(0.5)  # give some time to the server
        retries -= 1
    status_code = r.status_code
    if not bibtex.startswith('@'):
        raise ValueError('Query %(url)s for BibTeX for a DOI %(doi)s (wrong doi?) has failed. '
                         'Response code %(status_code)d. '
                         #'BibTeX response was: %(bibtex)s'
                         % locals())
    if not exists(cached):
        cache_dir = dirname(cached)
        if not exists(cache_dir):
            os.makedirs(cache_dir)
        with open(cached, 'w') as f:
            if PY2:
                f.write(bibtex.encode('utf-8'))
            else:
                f.write(bibtex)
    return bibtex


class EnumeratedEntries(Iterator):
    """A container of entries enumerated referenced by their entry_key"""
    def __init__(self):
        self._keys2refnr = {}
        self._refnr2keys = {}
        self._refnr = 1

    def add(self, entry_key):
        """Add entry_key and update refnr"""
        if entry_key not in self._keys2refnr:
            self._keys2refnr[entry_key] = self._refnr
            self._refnr2keys[self._refnr] = entry_key
            self._refnr += 1

    def __getitem__(self, item):
        if item not in self._keys2refnr:
            raise KeyError('{0} not present'.format(item))
        return self._keys2refnr[item]

    def fromrefnr(self, refnr):
        if refnr not in self._refnr2keys:
            raise KeyError('{0} not present'.format(refnr))
        return self._refnr2keys[refnr]

    def __iter__(self):
        return iteritems(self._keys2refnr)

    # Python 3
    def __next__(self):
        return self.next()

    def next(self):
        yield next(self.__iter__())

    def __len__(self):
        return len(self._keys2refnr)


class TextOutput(object):  # TODO some parent class to do what...?
    def __init__(self, fd, collector, style=None):
        self.fd = fd
        self.collector = collector
        # TODO: check that CLS style actually exists
        self.style = style
        if 'DUECREDIT_STYLE' in os.environ.keys():
            self.style = os.environ['DUECREDIT_STYLE']
        else:
            self.style = 'harvard1'

    # TODO: refactor name to sth more intuitive
    def _model_citations(self, tags=None):
        if not tags:
            tags = os.environ.get('DUECREDIT_REPORT_TAGS', 'reference-implementation,implementation').split(',')
        tags = set(tags)

        citations = self.collector.citations
        if tags != {'*'}:
            # Filter out citations
            citations = dict((k, c)
                             for k, c in iteritems(citations)
                             if tags.intersection(c.tags))

        packages = {}
        modules = {}
        objects = {}

        for key in ('citations', 'entry_keys'):
            packages[key] = defaultdict(list)
            modules[key] = defaultdict(list)
            objects[key] = defaultdict(list)

        # for each path store both a list of entry keys and of citations
        for (path, entry_key), citation in iteritems(citations):
            if ':' in path:
                target_dict = objects
            elif '.' in path:
                target_dict = modules
            else:
                target_dict = packages
            target_dict['citations'][path].append(citation)
            target_dict['entry_keys'][path].append(entry_key)
        return packages, modules, objects

    def dump(self, tags=None):
        # get 'model' of citations
        packages, modules, objects = self._model_citations(tags)
        # mapping key -> refnr
        enum_entries = EnumeratedEntries()

        citations_ordered = []
        # set up view

        # package level
        sublevels = [modules, objects]
        for package in sorted(packages['entry_keys']):
            for entry_key in packages['entry_keys'][package]:
                enum_entries.add(entry_key)
            citations_ordered.append(package)
            # sublevels
            for sublevel in sublevels:
                for obj in sorted(filter(lambda x: package in x, sublevel['entry_keys'])):
                    for entry_key_obj in sublevel['entry_keys'][obj]:
                        enum_entries.add(entry_key_obj)
                    citations_ordered.append(obj)

        # Now we can "render" different views of our "model"
        # Here for now just text BUT that is where we can "split" the logic and provide
        # different renderings given the model -- text, rest, md, tex+latex, whatever
        self.fd.write('\nDueCredit Report:\n')

        for path in citations_ordered:
            if ':' in path:
                self.fd.write('  ')
                target_dict = objects
            elif '.' in path:
                self.fd.write('  ')
                target_dict = modules
            else:
                target_dict = packages
            # TODO: absorb common logic into a common function
            citations = target_dict['citations'][path]
            entry_keys = target_dict['entry_keys'][path]
            descriptions = sorted(map(str, set(str(r.description) for r in citations)))
            versions = sorted(map(str, set(str(r.version) for r in citations)))
            refnrs = sorted([str(enum_entries[entry_key]) for entry_key in entry_keys])
            self.fd.write('- {0} / {1} (v {2}) [{3}]\n'.format(
                ", ".join(descriptions), path, ', '.join(versions), ', '.join(refnrs)))

        # Print out some stats
        obj_names = ('packages', 'modules', 'functions')
        n_citations = map(len, (packages['citations'], modules['citations'], objects['citations']))
        for citation_type, n in zip(obj_names, n_citations):
            self.fd.write('\n{0} {1} cited'.format(n, citation_type))

        if enum_entries:
            citations_fromentrykey = self.collector._citations_fromentrykey()
            self.fd.write('\n\nReferences\n' + '-' * 10 + '\n')
            # collect all the entries used
            refnr_key = [(nr, enum_entries.fromrefnr(nr)) for nr in range(1, len(enum_entries)+1)]
            for nr, key in refnr_key:
                self.fd.write('\n[{0}] '.format(nr))
                self.fd.write(get_text_rendering(citations_fromentrykey[key], style=self.style))
            self.fd.write('\n')

def get_text_rendering(citation, style='harvard1'):
    from .collector import Citation
    entry = citation.entry
    if isinstance(entry, Doi):
        bibtex_rendering = get_bibtex_rendering(entry)
        bibtex_citation = copy.copy(citation)
        bibtex_citation.set_entry(bibtex_rendering)
        return get_text_rendering(bibtex_citation)
    elif isinstance(entry, BibTeX):
        return format_bibtex(entry, style=style)
    else:
        return str(entry)


def get_bibtex_rendering(entry):
    if isinstance(entry, Doi):
        return BibTeX(import_doi(entry.doi))
    elif isinstance(entry, BibTeX):
        return entry
    else:
        raise ValueError("Have no clue how to get bibtex out of %s" % entry)


def format_bibtex(bibtex_entry, style='harvard1'):
    try:
        from citeproc.source.bibtex import BibTeX as cpBibTeX
        import citeproc as cp
    except ImportError as e:
        raise RuntimeError(
            "For formatted output we need citeproc and all of its dependencies "
            "(such as lxml) but there is a problem while importing citeproc: %s"
            % str(e))
    key = bibtex_entry.get_key()
    # need to save it temporarily to use citeproc-py
    fname = tempfile.mktemp(suffix='.bib')
    try:
        with open(fname, 'wt') as f:
            bibtex = bibtex_entry.rawentry
            bibtex = bibtex.replace(u'\u2013', '--') + "\n"
            # TODO: manage to save/use UTF-8
            if PY2:
                bibtex = bibtex.encode('ascii', 'ignore')
            f.write(bibtex)
        # We need to avoid cpBibTex spitting out warnings
        old_filters = warnings.filters[:]  # store a copy of filters
        warnings.simplefilter('ignore', UserWarning)
        try:
            bib_source = cpBibTeX(fname)
        except Exception as e:
            lgr.error("Failed to process BibTeX file %s: %s" % (fname, e))
            return "ERRORED: %s" % str(e)
        finally:
            # return warnings back
            warnings.filters = old_filters
        bib_style = cp.CitationStylesStyle(style, validate=False)
        # TODO: specify which tags of formatter we want
        bibliography = cp.CitationStylesBibliography(bib_style, bib_source,
                                                     cp.formatter.plain)
        citation = cp.Citation([cp.CitationItem(key)])
        bibliography.register(citation)
    finally:
        if not os.environ.get("DUECREDIT_KEEPTEMP"):
            os.unlink(fname)

    biblio_out = bibliography.bibliography()
    assert(len(biblio_out) == 1)
    biblio_out = ''.join(biblio_out[0])
    return biblio_out # if biblio_out else str(bibtex_entry)

# TODO: harmonize order of arguments
class PickleOutput(object):
    def __init__(self, collector, fn=DUECREDIT_FILE):
        self.collector = collector
        self.fn = fn

    def dump(self):
        with open(self.fn, 'wb') as f:
            pickle.dump(self.collector, f)

    @classmethod
    def load(cls, filename=DUECREDIT_FILE):
        with open(filename, 'rb') as f:
            return pickle.load(f)

class BibTeXOutput(object):  # TODO some parent class to do what...?
    def __init__(self, fd, collector):
        self.fd = fd
        self.collector = collector

    def dump(self):
        for citation in self.collector.citations.values():
            try:
                bibtex = get_bibtex_rendering(citation.entry)
            except:
                lgr.warning("Failed to generate bibtex for %s" % citation.entry)
                continue
            self.fd.write(bibtex.rawentry + "\n")


def load_due(filename):
    return PickleOutput.load(filename)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Citation and citations Collector classes"""

import os
import sys
from functools import wraps

from six import iteritems, itervalues

from .config import DUECREDIT_FILE
from .entries import DueCreditEntry
from .stub import InactiveDueCreditCollector
from .io import TextOutput, PickleOutput
from .utils import never_fail, borrowdoc
from .versions import external_versions
from collections import namedtuple

import logging
lgr = logging.getLogger('duecredit.collector')

CitationKey = namedtuple('CitationKey', ['path', 'entry_key'])

class Citation(object):
    """Encapsulates citations and information on their use"""

    def __init__(self, entry, description=None, path=None, version=None,
                 cite_module=False, tags=['implementation']):
        """Cite a reference

        Parameters
        ----------
        entry: str or DueCreditEntry
          The entry to use, either identified by its id or a new one (to be added)
        description: str, optional
          Description of what this functionality provides
        path: str
          Path to the object which this citation associated with.  Format is
          "module[.submodules][:[class.]method]", i.e. ":" is used to separate module
          path from the path within the module.
        version: str or tuple, version
          Version of the beast (e.g. of the module) where applicable
        cite_module: bool, optional
          If it is a module citation, setting it to True would make that module citeable
          even without internal duecredited functionality invoked.  Should be used only for
          core packages whenever it is reasonable to assume that its import constitute
          its use (e.g. numpy)
        tags: list of str, optional
          Tags to associate with the given code/reference combination.  Some tags have
          associated semantics in duecredit, e.g.
          - "implementation" [default] tag describes as an implementation of the cited
            method
          - "reference-implementation" tag describes as the original implementation (ideally
            by the authors of the paper) of the cited method
          - "another-implementation" tag describes some other implementation of the method
          - "use" tag points to publications demonstrating a worthwhile noting use
            the method
          - "edu" references to tutorials, textbooks and other materials useful to learn
            more
          - "donate" should be commonly used with Url entries to point to the websites
            describing how to contribute some funds to the referenced project
        """
        if path is None:
            raise ValueError('Must specify path')
        self._entry = entry
        self._description = description
        # We might want extract all the relevant functionality into a separate class
        self._path = path
        self._cite_module = cite_module
        self.tags = tags or []
        self.version = version
        self.count = 0

    def __repr__(self):
        args = [repr(self._entry)]
        if self._description:
            args.append("description={0}".format(repr(self._description)))
        if self._path:
            args.append("path={0}".format(repr(self._path)))
        if self._cite_module:
            args.append("cite_module={0}".format(repr(self._cite_module)))

        if args:
            args = ", ".join(args)
        else:
            args = ""
        return self.__class__.__name__ + '({0})'.format(args)

    @property
    def path(self):
        return self._path

    @property
    def cite_module(self):
        return self._cite_module

    @path.setter
    def path(self, path):
        # TODO: verify value, if we are not up for it -- just make _path public
        self._path = path

    @property
    def entry(self):
        return self._entry

    @property
    def description(self):
        return self._description

    @property
    def cites_module(self):
        if not self.path:
            return None
        return not (':' in self.path)

    @property
    def module(self):
        if not self.path:
            return None
        return self.path.split(':', 1)[0]

    @property
    def package(self):
        module = self.module
        if not module:
            return None
        return module.split('.', 1)[0]

    @property
    def objname(self):
        if not self.path:
            return None
        spl = self.path.split(':', 1)
        if len(spl) > 1:
            return spl[1]
        else:
            return None

    def __contains__(self, entry):
        """Checks if provided entry 'contained' in this one given its path

        If current entry is associated with a module, contained will be an entry
        of
        - the same module
        - submodule of the current module or function within

        If current entry is associated with a specific function/class, it can contain
        another entry if it really contains it as an attribute
        """
        if self.cites_module:
            return ((self.path == entry.path) or
                    (entry.path.startswith(self.path + '.')) or
                    (entry.path.startswith(self.path + ':')))
        else:
            return entry.path.startswith(self.path + '.')

    @property
    def key(self):
        return CitationKey(self.path, self.entry.get_key())

    @staticmethod
    def get_key(path, entry_key):
        return CitationKey(path, entry_key)

    def set_entry(self, newentry):
        self._entry = newentry


class DueCreditCollector(object):
    """Collect the references

    The mighty beast which will might become later a proxy on the way to
    talk to a real collector

    Parameters
    ----------
    entries : list of DueCreditEntry, optional
      List of reference items (BibTeX, Doi, etc) known to the collector
    citations : list of Citation, optional
      List of citations -- associations between references and particular
      code, with a description for its use, tags etc
    """

    # TODO?  rename "entries" to "references"?  or "references" is closer to "citations"
    def __init__(self, entries=None, citations=None):
        self._entries = entries or {}
        self.citations = citations or {}

    @never_fail
    def add(self, entry):
        """entry should be a DueCreditEntry object"""
        if isinstance(entry, list):
            for e in entry:
                self.add(e)
        else:
            key = entry.get_key()
            self._entries[key] = entry
            lgr.log(1, "Collector added entry %s", key)

    @never_fail
    def load(self, src):
        """Loads references from a file or other recognizable source

        ATM supported only
        - .bib files
        """
        # raise NotImplementedError
        if isinstance(src, str):
            if src.endswith('.bib'):
                self._load_bib(src)
            else:
                raise NotImplementedError('Format not yet supported')
        else:
            raise ValueError('Must be a string')

    def _load_bib(self, src):
        lgr.debug("Loading %s" % src)

    # # TODO: figure out what would be the optimal use for the __call__
    # def __call__(self, *args, **kwargs):
    #     # TODO: how to determine and cite originating module???
    #     #       - we could use inspect but many people complain
    #     #         that it might not work with other Python
    #     #         implementations
    #     pass # raise NotImplementedError

    @never_fail
    @borrowdoc(Citation, "__init__")
    def cite(self, entry, **kwargs):
        # TODO: if cite is invoked but no path is provided -- we must figure it out
        # I guess from traceback, otherwise how would we know later to associate it
        # with modules???
        path = kwargs.get('path', None)
        if path is None:
            raise ValueError('path must be provided')

        if isinstance(entry, DueCreditEntry):
            # new one -- add it
            self.add(entry)
            entry_ = self._entries[entry.get_key()]
        else:
            entry_ = self._entries[entry]

        entry_key = entry_.get_key()
        citation_key = Citation.get_key(path=path, entry_key=entry_key)
        try:
            citation = self.citations[citation_key]
        except KeyError:
            self.citations[citation_key] = citation = Citation(entry_, **kwargs)
        assert(citation.key == citation_key)
        # update citation count
        citation.count += 1

        # TODO: theoretically version shouldn't differ if we don't preload previous results
        if not citation.version:
            version = kwargs.get('version', None)

            if not version and citation.path:
                modname = citation.path.split('.', 1)[0]

                if '.' in modname:
                    package = modname.split('.', 1)[0]
                else:
                    package = modname

                # package_loaded = sys.modules.get(package)
                # if package_loaded:
                #     # find the citation for that module
                #     for citation in itervalues(self.citations):
                #         if citation.package == package \
                #                 and not citation.version:
                version = external_versions[package]
            citation.version = version

        return citation

    def _citations_fromentrykey(self):
        """Return a dictionary with the current citations indexed by the entry key"""
        citations_key = dict()
        for (path, entry_key), citation in iteritems(self.citations):
            if entry_key not in citations_key:
                citations_key[entry_key] = citation

        return citations_key


    @staticmethod
    def _args_match_conditions(conditions, *fargs, **fkwargs):
        """Helper to identify when to trigger citation given parameters to the function call
        """
        for (argpos, kwarg), values in iteritems(conditions):
            # main logic -- assess default and if get to the next one if
            # given argument is not present
            if not ((len(fargs) > argpos) or (kwarg in fkwargs)):
                if not ('DC_DEFAULT' in values):
                    # if value was specified but not provided and not default
                    # conditions are not satisfied
                    return False
                continue

            # "extract" the value.  Must be defined here
            value = "__duecredit_magical_undefined__"
            if len(fargs) > argpos:
                value = fargs[argpos]
            if kwarg in fkwargs:
                value = fkwargs[kwarg]
            assert(value != "__duecredit_magical_undefined__")

            if '.' in kwarg:
                # we were requested to condition based on the value of the attribute
                # of the value.  So get to the attribute(s) value
                for attr in kwarg.split('.')[1:]:
                    value = getattr(value, attr)

            # Value is present but not matching
            if not (value in values):
                return False

        # if checks passed -- we must have matched conditions
        return True

    @never_fail
    @borrowdoc(Citation, "__init__", replace="PLUGDOCSTRING")
    def dcite(self, *args, **kwargs):
        """Decorator for references.  PLUGDOCSTRING

        Parameters
        ----------
        conditions: dict, optional
          If reference should be cited whenever parameters to the function call
          satisfy given values (all of the specified).
          Each key in the dictionary is a 2 element tuple with first element, integer,
          pointing to a position of the argument in the original function call signature,
          while second provides the name, thus if used as a keyword argument.
          Use "DC_DEFAULT" keyword as a value to depict default value (e.g. if no
          explicit value was provided for that positional or keyword argument).
          If "keyword argument" is of the form "obj.attr1.attr2", then actual value
          for comparison would be taken by extracting attr1 (and then attr2) attributes
          from the provided value.  So, if desired to condition of the state of the object,
          you can use `(0, "self.attr1") : {...values...}`

        Examples
        --------

        >>> from duecredit import due
        >>> @due.dcite('XXX00', description="Provides an answer for meaningless existence")
        ... def purpose_of_life():
        ...     return None

        Conditional citation given argument to the function

        >>> @due.dcite('XXX00', description="Relief through the movement",
        ...            conditions={(1, 'method'): {'purge', 'DC_DEFAULT'}})
        ... @due.dcite('XXX01', description="Relief through the drug treatment",
        ...            conditions={(1, 'method'): {'drug'}})
        ... def relief(x, method='purge'):
        ...     if method == 'purge': return "crap"
        ...     elif method == 'drug': return "swallow"
        >>> relief("doesn't matter")
        'crap'

        Conditional based on the state of the object

        >>> class Citeable(object):
        ...     def __init__(self, param=None):
        ...         self.param = param
        ...     @due.dcite('XXX00', description="The same good old relief",
        ...                conditions={(0, 'self.param'): {'magic'}})
        ...     def __call__(self, data):
        ...         return sum(data)
        >>> Citeable('magic')([1, 2])
        3
        """
        def func_wrapper(func):
            conditions = kwargs.pop('conditions', {})
            path = kwargs.get('path', None)
            if not path:
                # deduce path from the actual function which was decorated
                # TODO: must include class name  but can't !!!???
                modname = func.__module__
                path = kwargs['path'] = '%s:%s' % (modname, func.__name__)
            else:
                # TODO: we indeed need to separate path logic outside
                modname = path.split(':', 1)[0]

            # if decorated function was invoked, it means that we need
            # to cite that even if it is a module. But do not override
            # value if user explicitely stated
            if 'cite_module' not in kwargs:
                kwargs['cite_module'] = True

            # TODO: might make use of inspect.getmro
            # see e.g.
            # http://stackoverflow.com/questions/961048/get-class-that-defined-method
            lgr.debug("Decorating func %s within module %s" % (func.__name__, modname))
            # TODO: unittest for all the __version__ madness

            # TODO: check if we better use wrapt module which provides superior "correctness"
            #       of decorating.  vcrpy uses wrapt, and that thing seems to wrap
            @wraps(func)
            def cite_wrapper(*fargs, **fkwargs):
                try:
                    if not conditions \
                            or self._args_match_conditions(conditions, *fargs, **fkwargs):
                        citation = self.cite(*args, **kwargs)
                except Exception as e:
                    lgr.warning("Failed to cite due to %s" % (e,))
                return func(*fargs, **fkwargs)

            cite_wrapper.__duecredited__ = func
            return cite_wrapper
        return func_wrapper

    @never_fail
    def __repr__(self):
        args = []
        if self.citations:
            args.append("citations={0}".format(repr(self.citations)))
        if self._entries:
            args.append("entries={0}".format(repr(self._entries)))

        if args:
            args = ", ".join(args)
        else:
            args = ""
        return self.__class__.__name__ + '({0})'.format(args)

    @never_fail
    def __str__(self):
        return self.__class__.__name__ + \
            ' {0:d} entries, {1:d} citations'.format(
                len(self._entries), len(self.citations))


# TODO: redo heavily -- got very messy
class CollectorSummary(object):
    """A helper which would take care about exporting citations upon its Death
    """
    def __init__(self, collector, outputs="stdout,pickle", fn=DUECREDIT_FILE):
        self._due = collector
        self.fn = fn
        # for now decide on output "format" right here
        self._outputs = [
            self._get_output_handler(
                type_.lower().strip(), collector, fn=fn)
            for type_ in os.environ.get('DUECREDIT_OUTPUTS', outputs).split(',')
            if type_
        ]

    @staticmethod
    def _get_output_handler(type_, collector, fn=None):
        # just a little factory
        if type_ in ("stdout", "stderr"):
            return TextOutput(getattr(sys, type_), collector)
        elif type_ == "pickle":
            return PickleOutput(collector, fn=fn)
        else:
            raise NotImplementedError()

    def dump(self):
        for output in self._outputs:
            output.dump()


# TODO:  provide HTML, MD, RST etc formattings




# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Automatic injection of bibliography entries for nipy module
"""

from ..entries import Doi

# If defined, would determine from which to which version of the corresponding
# module to care about
min_version = None
max_version = None


def inject(injector):
    injector.add('nipy', None, Doi('10.1016/S1053-8119(09)72223-2'),
                 description="Library fMRI data analysis",
                 tags=['implementation'])

    for f, d in [('spectral_decomposition', 'PCA decomposition of symbolic HRF shifted over time'),
                 ('taylor_approx', 'A Taylor series approximation of an HRF shifted over time')]:
        injector.add('nipy.modalities.fmri.fmristat.hrf', f, Doi('10.1006/nimg.2002.1096'),
            description=d, tags=['implementation'])


# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

"""
Automatic injection of bibliography entries for mdp module
"""


from ..entries import Doi, BibTeX, Url

# If defined, would determine from which to which version of the corresponding
# module to care about
min_version = None
max_version = None


def inject(injector):
    injector.add('mdp', None, Doi('10.3389/neuro.11.008.2008'),
                 description="Modular toolkit for Data Processing (MDP): a Python data processing framework",
                 tags=['implementation'])

    injector.add('mdp.nodes', 'PCANode.train', Doi('10.1007/b98835'),
                 description="Principal Component Analysis (and filtering)",
                 tags=['implementation'])

    injector.add('mdp.nodes', 'NIPALSNode.train', BibTeX("""
        @incollection{Word1966,
            author={Wold, H.},
            title={Nonlinear estimation by iterative least squares procedures.},
            booktitle={Research Papers in Statistics},
            publisher={Wiley}
            year={1966},
            editor={David, F.},
            pages={411--444},
            }
        """), description="Principal Component Analysis using the NIPALS algorithm.",
        tags=['edu'])

    injector.add('mdp.nodes', 'FastICANode.train', Doi('10.1109/72.761722'),
                 description="Independent Component Analysis using the FastICA algorithm",
                 tags=['implementation'])

    injector.add('mdp.nodes', 'CuBICANode.train', Doi('10.1109/TSP.2004.826173'),
                 description='Independent Component Analysis using the CuBICA algorithm.',
                 tags=['implementation'])
    
    injector.add('mdp.nodes', 'NIPALSNode.train', BibTeX("""
        @conference{ZieheMuller1998,
            author={Ziehe, Andreas and Muller, Klaus-Robert},
            title={TDSEP an efficient algorithm for blind separation using time structure.},
            booktitle={Proc. 8th Int. Conf. Artificial Neural Networks},
            year={1998},
            editor={Niklasson, L, Boden, M, and Ziemke, T},
            publisher={ICANN}
            }
        """), description='Independent Component Analysis using the TDSEP algorithm',
        tags=['edu'])


    injector.add('mdp.nodes', 'JADENode.train', Doi('10.1049/ip-f-2.1993.0054'),
                    description='Independent Component Analysis using the JADE algorithm',
                    tags=['implementation'])
    injector.add('mdp.nodes', 'JADENode.train', Doi('10.1162/089976699300016863'),
                    description='Independent Component Analysis using the JADE algorithm',
                    tags=['implementation'])

    injector.add('mdp.nodes', 'SFANode.train', Doi('10.1162/089976602317318938'),
                    description='Slow Feature Analysis',
                    tags=['implementation'])

    injector.add('mdp.nodes', 'SFA2Node.train', Doi('10.1162/089976602317318938'),
                    description='Slow Feature Analysis (via the space of inhomogeneous polynomials)',
                    tags=['implementation'])
    

    injector.add('mdp.nodes', 'ISFANode.train', Doi('10.1007/978-3-540-30110-3_94'),
                    description='Independent Slow Feature Analysis',
                    tags=['implementation'])

 
    injector.add('mdp.nodes', 'XSFANode.train', BibTeX("""
        @article{SprekelerZitoWiskott2009,
            author={Sprekeler, H., Zito, T., and Wiskott, L.},
            title={An Extension of Slow Feature Analysis for Nonlinear Blind Source Separation.},
            journal={Journal of Machine Learning Research.},
            year={2009},
            volume={15},
            pages={921--947},
            }
        """), description="Non-linear Blind Source Separation using Slow Feature Analysis",
        tags=['edu'])

    injector.add('mdp.nodes', 'FDANode.train', BibTeX("""
        @book{Bishop2011,
            author={Bishop, Christopher M.},
            title={Neural Networks for Pattern Recognition},
            publisher={Oxford University Press, Inc}
            year={2011},
            pages={105--112},
            }
        """), description="(generalized) Fisher Discriminant Analysis",
        tags=['edu'])

    # etc...



#!/usr/bin/env python
"""
duecredit -- publications (donations, etc) tracer
"""

import re
import os
import sys
import re

from datetime import datetime
from setuptools import setup
from pkgutil import walk_packages
from subprocess import Popen, PIPE
from os.path import exists

# Adopted from citeproc-py
#  License: BSD-2
#  Copyright 2011-2013 Brecht Machiels

PACKAGE = 'duecredit'
PACKAGE_ABSPATH = os.path.abspath(PACKAGE)
VERSION_FILE = PACKAGE + '/version.py'

# retrieve the version number from git or VERSION_FILE
# inspired by http://dcreager.net/2010/02/10/setuptools-git-version-numbers/

try:
    if exists('debian/copyright'):
        print('Generating version.py out of debian/copyright information')
        # building debian package. Deduce version from debian/copyright
        with open('debian/changelog', 'r') as f:
            lines = f.readlines()
        __version__ = re.sub('(.*)-(.*?)$', r'\1.debian\2',
                             lines[0].split()[1].strip('()')
                             ).replace('-', '.')
        # TODO: unify format whenever really bored ;)
        __release_date__ = re.sub('^ -- .*>\s*(.*)', r'\1',
                                  list(filter(lambda x: x.startswith(' -- '), lines))[0].rstrip())
    else:
        print('Attempting to get version number from git...')
        git = Popen(['git', 'describe', '--abbrev=4', '--dirty'],
                    stdout=PIPE, stderr=sys.stderr)
        if git.wait() != 0:
            raise OSError
        line = git.stdout.readlines()[0]
        __version__ = line.strip().decode('ascii')
        __release_date__ = datetime.now().strftime('%b %d %Y, %H:%M:%S')
    with open(VERSION_FILE, 'w') as version_file:
        version_file.write("__version__ = '{0}'\n".format(__version__))
        version_file.write("__release_date__ = '{0}'\n".format(__release_date__))
except OSError as e:
    print('Assume we are running from a source distribution.')
    # read version from VERSION_FILE
    if os.path.exists(VERSION_FILE):
        with open(VERSION_FILE) as version_file:
            code = compile(version_file.read(), VERSION_FILE, 'exec')
            exec(code)
    else:
        __version__ = '0.unknown'

with open('README.md') as file:
    README = file.read()

def find_packages(path, prefix):
    yield prefix
    prefix = prefix + "."
    for _, name, ispkg in walk_packages(path, prefix):
        if ispkg:
            yield name


setup(
    name=PACKAGE,
    version=__version__,
    packages=list(find_packages([PACKAGE_ABSPATH], PACKAGE)),
    # package_data={
    #     PACKAGE: [
    #         'tests/envs/nolxml/lxml.py',
    #         'tests/envs/stubbed/README.txt',
    #         'tests/envs/stubbed/due.py',
    #         'tests/envs/stubbed/script.py',
    #     ]
    # },
    scripts=[],
    install_requires=['requests', 'citeproc-py'],
    include_package_data=True,
    provides=[PACKAGE],
    #test_suite='nose.collector',
    entry_points={
        'console_scripts': [
             'duecredit=duecredit.cmdline.main:main',
        ],
    },
    author='Yaroslav Halchenko, Matteo Visconti di Oleggio Castello',
    author_email='yoh@onerussian.com',
    description='Publications (and donations) tracer',
    long_description="""\
duecredit is being conceived to address the problem of inadequate
citation of scientific software and methods, and limited visibility of
donation requests for open-source software.

It provides a simple framework (at the moment for Python only) to
embed publication or other references in the original code so they are
automatically collected and reported to the user at the necessary
level of reference detail, i.e. only references for actually used
functionality will be presented back if software provides multiple
citeable implementations.

To get a sense of what duecredit is about, run for example shipped along
example script, or your analysis script with `-m duecredit`, e.g.

    python -m duecredit examples/example_scipy.py

""",
    url='https://github.com/duecredit/duecredit',
    # Download URL will point to the latest release, thus suffixes removed
    download_url='https://github.com/duecredit/duecredit/releases/tag/%s' % re.sub('-.*$', '', __version__),
    keywords=['citation tracing',],
    license='2-clause BSD License',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Environment :: Other Environment',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Legal Industry',
        'Intended Audience :: Other Audience',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Topic :: Documentation',
        'Topic :: Printing',
        'Topic :: Software Development :: Documentation',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""Module to help maintain a registry of versions for external modules etc
"""
import sys
from os import linesep
from six import string_types

from distutils.version import StrictVersion, LooseVersion

# To depict an unknown version, which can't be compared by mistake etc
class UnknownVersion:
    """For internal use
    """

    def __str__(self):
        return "UNKNOWN"

    def __cmp__(self, other):
        if other is self:
            return 0
        raise TypeError("UNKNOWN version is not comparable")


class ExternalVersions(object):
    """Helper to figure out/use versions of the external modules.

    It maintains a dictionary of `distuil.version.StrictVersion`s to make
    comparisons easy.  If version string doesn't conform the StrictVersion
    LooseVersion will be used.  If version can't be deduced for the module,
    'None' is assigned
    """

    UNKNOWN = UnknownVersion()

    def __init__(self):
        self._versions = {}

    @classmethod
    def _deduce_version(klass, module):
        version = None
        for attr in ('__version__', 'version'):
            if hasattr(module, attr):
                version = getattr(module, attr)
                break

        if isinstance(version, tuple) or isinstance(version, list):
            #  Generate string representation
            version = ".".join(str(x) for x in version)

        if version:
            try:
                return StrictVersion(version)
            except ValueError:
                # let's then go with Loose one
                return LooseVersion(version)
        else:
            return klass.UNKNOWN

    def __getitem__(self, module):
        # when ran straight in its source code -- fails to discover nipy's version.. TODO
        #if module == 'nipy':
        #    import pdb; pdb.set_trace()
        if not isinstance(module, string_types):
            modname = module.__name__
        else:
            modname = module
            module = None

        if modname not in self._versions:
            if module is None:
                if modname not in sys.modules:
                    try:
                        module = __import__(modname)
                    except ImportError:
                        return None
                else:
                    module = sys.modules[modname]

            self._versions[modname] = self._deduce_version(module)

        return self._versions.get(modname, self.UNKNOWN)

    def keys(self):
        """Return names of the known modules"""
        return self._versions.keys()

    def __contains__(self, item):
        return item in self._versions

    @property
    def versions(self):
        """Return dictionary (copy) of versions"""
        return self._versions.copy()

    def dumps(self, indent=False, preamble="Versions:"):
        """Return listing of versions as a string

        Parameters
        ----------
        indent: bool or str, optional
          If set would instruct on how to indent entries (if just True, ' '
          is used). Otherwise returned in a single line
        preamble: str, optional
          What preamble to the listing to use
        """
        if indent and (indent is True):
            indent = ' '
        items = ["%s=%s" % (k, self._versions[k]) for k in sorted(self._versions)]
        out = "%s" % preamble
        if indent:
            out += (linesep + indent).join([''] + items) + linesep
        else:
            out += " " + ' '.join(items)
        return out


external_versions = ExternalVersions()



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 noet:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the duecredit package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##

import sys

from mock import patch
from six.moves import StringIO
from nose.tools import assert_raises, assert_equal

from .. import __version__
from ..cmdline import main

def test_import():
    import duecredit.cmdline
    import duecredit.cmdline.main

@patch('sys.stdout', new_callable=StringIO)
def test_main_help(stdout):
    assert_raises(SystemExit, main.main, ['--help'])
    assert(stdout.getvalue().lstrip().startswith("Usage: "))

# differs among Python versions -- catch both
@patch('sys.std' + ('err' if sys.version_info < (3, 4) else 'out'), new_callable=StringIO)
def test_main_version(out):
    assert_raises(SystemExit, main.main, ['--version'])
    assert_equal((out.getvalue()).split('\n')[0], "duecredit %s" % __version__)

# smoke test the cmd_summary
# TODO: carry sample .duecredit.p, point to that file, mock TextOutput and BibTeXOutput .dumps
def test_smoke_cmd_summary():
    main.main(['summary'])

# test the not implemented cmd_test
def test_cmd_test():
    assert_raises(SystemExit, main.main, ['test'])